/*
This software and its source code is in the Public Domain.
Use at your own risk.
*/

#include <PalmOS.h>
#include <PceNativeCall.h>
#include "FixWiFi.h"
#include "FixWiFi_Res.h"

UInt16 launchCode;

typedef struct {
    Boolean active;
} PreferenceType;


MemHandle panelIDsH = NULL;
SysDBListItemType* panelData;
Char** panelNames = NULL;
UInt16 panelCount = 0;


static void InstallResident( Boolean state );


PreferenceType pref;


UInt16 GetObjectIndex( UInt16 id )
{
    return FrmGetObjectIndex( FrmGetActiveForm(), id );
}


void* GetObjectPtr( UInt16 id )
{
    return FrmGetObjectPtr( FrmGetActiveForm(), GetObjectIndex( id ) );
}



void Show( UInt16 id, Boolean state )
{
    UInt16 index = GetObjectIndex( id );

    if ( state )
        FrmShowObject( FrmGetActiveForm(), index );
    else
        FrmHideObject( FrmGetActiveForm(), index );
    if ( FrmGetObjectType( FrmGetActiveForm(), index ) == frmControlObj )
        CtlSetUsable( GetObjectPtr( id ), state);
}


void LoadPrefs0( PreferenceType* p )
{
    UInt16 size;

    MemSet( p, sizeof( *p ), 0 );

    p->active = true;

    size = 0;

    if ( noPreferenceFound ==
             PrefGetAppPreferences( appCreator, PREF_ID_MAIN, NULL, &size, true ) ) {
        return;
    }

    if ( sizeof( *p ) < size ) {
        size = sizeof( *p );
    }

    PrefGetAppPreferences( appCreator, PREF_ID_MAIN, p, &size, true );
}



void SavePrefs0( PreferenceType* p )
{
    PrefSetAppPreferences( appCreator, PREF_ID_MAIN, 0, p,
        sizeof( PreferenceType ), true );
}



#ifndef APP
void ClearPanelList( void )
{
    if ( 0 < panelCount ) {
        MemPtrFree( panelNames );
        MemHandleUnlock( panelIDsH );
        MemHandleFree( panelIDsH );

        panelNames = NULL;
        panelCount = 0;
    }
}



void FixObjectWidthRight( UInt16 id, Coord width )
{
    RectangleType bounds;
    
    FrmGetObjectBounds( FrmGetActiveForm(), GetObjectIndex( id ), &bounds );

    bounds.topLeft.x = bounds.topLeft.x + bounds.extent.x - width;
    bounds.extent.x = width;

    FrmSetObjectBounds( FrmGetActiveForm(), GetObjectIndex( id ), &bounds );
}



void InitPanelList( void )
{
    Int16     myPanel;
    Coord     maxWidth;

    ClearPanelList();
    
    myPanel  = -1;
    maxWidth = 0;

    if ( SysCreatePanelList( &panelCount, &panelIDsH ) &&
         0 < panelCount ) {
        panelData = MemHandleLock( panelIDsH );

        panelNames = MemPtrNew( panelCount * sizeof( Char* ) );

        if ( panelNames == NULL ) {
            panelCount = 0;
            MemHandleUnlock( panelIDsH );
            MemHandleFree( panelIDsH );
        }
        else {
            UInt16 i;
            UInt16 card;
            LocalID id;
            Err    err;

            FntSetFont( stdFont );

            err = SysCurAppDatabase( &card, &id );

            if ( err != errNone ) {
                card = 0xFFFF;
                id = NULL;
            }

            for ( i = 0 ; i < panelCount ; i++ ) {
                 Coord w;

                 panelNames[ i ] = panelData[ i ].name;
                 if ( panelData[ i ].dbID == id )
                     myPanel = i;
                 w = FntCharsWidth( panelNames[ i ], StrLen( panelNames[ i ] ) );

                 if ( maxWidth < w )
                     maxWidth = w;
            }
        }
    }

    if ( 0 < panelCount ) {
        Coord selectionWidth;
        
        selectionWidth = 18;
        
        if ( 0 <= myPanel )
            selectionWidth += FntCharsWidth( panelNames[ myPanel ], 
                                  StrLen( panelNames[ myPanel ] ) );

        LstSetListChoices( GetObjectPtr( lstPanelList ), panelNames, panelCount );
        LstSetSelection( GetObjectPtr( lstPanelList ), myPanel );
        LstSetHeight( GetObjectPtr( lstPanelList ), panelCount );
        FixObjectWidthRight( trgPanelList, selectionWidth );
        FixObjectWidthRight( lstPanelList, maxWidth );
    }
    else {
        LstSetListChoices( GetObjectPtr( lstPanelList ), NULL, 0 );
    }
    if ( 0 <= myPanel ) {
        CtlSetLabel( GetObjectPtr( trgPanelList ), panelNames[ myPanel ] );
    }
    else {
        CtlSetLabel( GetObjectPtr( trgPanelList ), "" );
    }
}
#endif



void PreparePanel( void )
{
#ifdef APP
    Show( btnDone, false );
    Show( trgPanelList, false );
#else
    if ( launchCode == sysAppLaunchCmdPanelCalledFromApp ) {
        Show( btnDone, true );
        Show( trgPanelList, false );
        Show( lblTitle, true );
    }
    else {
        InitPanelList();

        Show( btnDone, false );
        Show( trgPanelList, true );
        Show( lblTitle, false );
    }

#endif
}



static Boolean Handler_frmMain( EventType* event )
{
    Boolean handled;
    handled = false;

    switch (event->eType) {
        case winEnterEvent:
            handled = false;
            break;

        case frmOpenEvent: {
            MemHandle handle;
            MemHandle infoHandle;

            CtlSetValue( GetObjectPtr( btnOn ), pref.active );
            CtlSetValue( GetObjectPtr( btnOff ), ! pref.active );

            infoHandle = DmGetResource( 'tSTR', strInfo );

            handle     = MemHandleNew( MemHandleSize( infoHandle ) );

            if ( handle != NULL ) {
                Char* p;
                Char* q;

                p = MemHandleLock( infoHandle );
                q = MemHandleLock( handle );

                MemMove( q, p, MemHandleSize( infoHandle ) );
                MemHandleUnlock( handle );
                MemHandleUnlock( infoHandle );
                DmReleaseResource( infoHandle );

                FldSetTextHandle( GetObjectPtr( fldInfo ), handle );
            }

            PreparePanel();

            FrmDrawForm( FrmGetActiveForm() );

            handled = true;
            break;
        }

        case popSelectEvent:
            if ( event->data.popSelect.selection != noListSelection ) {
                switch ( event->data.popSelect.controlID ) {
                    case trgPanelList: {
                        SysDBListItemType* item;

                        item = panelData + event->data.popSelect.selection;

                        SysUIAppSwitch( item->cardNo, item->dbID, 0, NULL );
                        
                        handled = true;
                        break;
                    }
                }
            }

        case ctlSelectEvent: {
            Boolean value;

            value = !! CtlGetValue( GetObjectPtr( event->data.ctlEnter.controlID ) );

            switch( event->data.ctlEnter.controlID ) {

                case btnOn:
                    if ( pref.active != value ) {
                        pref.active = value;
                        InstallResident( pref.active );
                        SavePrefs0( &pref );
                    }
                    break;
                case btnOff:
                    if ( pref.active != ! value ) {
                        pref.active = ! value;
                        InstallResident( pref.active );
                        SavePrefs0( &pref );
                    }
                    break;
                case btnDone: {
                    EventType e;
                    
                    MemZero( &e, sizeof( EventType ) );
                    e.eType = appStopEvent;
                    
                    EvtAddEventToQueue( &e );
                    break;
                }
            }
            break;
        }
        default:
            break;
    }

    return handled;
}



static void EventLoop( void )
{
    EventType event;
    Boolean   handled;

    do {
       	EvtGetEvent( &event, evtWaitForever );

       	handled = SysHandleEvent( &event );

       	if ( ! handled ) {
            switch ( event.eType ) {
                case frmLoadEvent: {
                    UInt16    formID;
                    FormType* form;

                    formID = event.data.frmLoad.formID;
                    form = FrmInitForm( formID );

                    FrmSetActiveForm( form );

                    switch ( formID ) {
                        case frmMain:
                            FrmSetEventHandler( form, Handler_frmMain );
                            break;
                    }
                    handled = true;
                }
                default:
                    handled = FrmDispatchEvent( &event );
            }
       	}
    } while ( event.eType != appStopEvent );

    FrmCloseAllForms();
}



static Boolean ApplicationStart( void )
{
    LoadPrefs0( &pref );

    InstallResident( pref.active );

    FrmGotoForm( frmMain );

    return true;
}



static void ApplicationStop( void )
{
#ifndef APP
    ClearPanelList();
#endif
}



static Boolean FindMe( UInt16* card, LocalID* id )
{
    DmSearchStateType state;

    return errNone == DmGetNextDatabaseByTypeCreator( true, &state,
               appType, appCreator,
               true, card, id );
}



static void CopyFromHandle( void* dest, MemHandle srcH, UInt32 size )
{
    void* src;

    if ( size == 0 )
        return;

    src = MemHandleLock( srcH );

    MemMove( dest, src, size );

    MemHandleUnlock( srcH );
}



static Err ArmSysNotifyRegister_hede( Err ( *f )( EventType* eventP ) )
{
     void* function;
     Err   value;
     MemHandle armH;

     armH = DmGetResource( 'armc', armNotify );

     if ( armH == NULL )
         return 1;

     function = MemHandleLock( armH );
     
     value = PceNativeCall( function, f );

     MemHandleUnlock( armH );
     DmReleaseResource( armH );

     return value;
}



static void InstallResident( Boolean state )
{
    UInt16  card;
    LocalID id;

    if ( ! FindMe( &card, &id ) )
        return;

    if ( state ) {
        SysNotifyRegister( card, id, 'WiDn', NULL, 0, NULL );
        SysNotifyRegister( card, id, appCreator, NULL, 0, NULL );
    }
    else {
        SysNotifyUnregister( card, id, 'WiDn', 0 );
        SysNotifyUnregister( card, id, appCreator, 0 );
    }
}



static void InstallHede( Boolean state )
{
    void* residentCode;
    LocalID id;
    UInt16  card;
    UInt32  tempValue;
    UInt32  processor;
    Boolean arm;

    if ( ! FindMe( &card, &id ) )
        return;

    arm = ( errNone == FtrGet( sysFileCSystem, sysFtrNumProcessorID, &processor ) &&
                   sysFtrNumProcessorIsARM( processor ) );
                   
//    arm = false;

    residentCode = ( void* )Get( FTR_RESIDENT_CODE );

    if ( residentCode == NULL && ! state )
        return;

    if ( state ) {
        MemHandle codeH;
        void* shadowedSegment;
        ResidentData*  data;
        UInt32 size;

        if ( residentCode != NULL )
            InstallResident( false );

        data  = MemPtrNew( sizeof( ResidentData ) );
        if ( data == NULL )
            return;

        MemPtrSetOwner( data, 0 );
        MemZero( data, sizeof( ResidentData ) );

        if ( arm )
            codeH = DmGetResource( 'armc', armResident );
        else
            codeH = DmGetResource( 'code', SECTION_ID_RESIDENT );

        if ( codeH == NULL ) {
            MemPtrFree( data );
            return;
        }

        size            = MemHandleSize( codeH );
        shadowedSegment = MemPtrNew( size );

        if ( shadowedSegment == NULL ) /* Better error handling? */ {
            MemPtrFree( data );
            return;
        }

        Set( FTR_RESIDENT_DATA, ( UInt32 )data );

        MemPtrSetOwner( shadowedSegment, 0 );
        CopyFromHandle( shadowedSegment, codeH, size );

        Set( FTR_RESIDENT_CODE, shadowedSegment );

        Set( FTR_TIMER_STOP, TimGetSeconds() + 2 );

        if ( arm )
            ArmSysNotifyRegister_hede( shadowedSegment );
        else
            SysNotifyRegister( card, id, sysNotifyEventDequeuedEvent,
                ( void* )shadowedSegment, 0, NULL );
    }
    else {
        if ( arm )
            ArmSysNotifyRegister_hede( NULL );
        else
            SysNotifyUnregister( card, id, sysNotifyEventDequeuedEvent, 0 );

        MemPtrFree( residentCode );
        Set( FTR_RESIDENT_CODE, 0 );
        MemPtrFree( ( void* )Get( FTR_RESIDENT_DATA ) );
        Set( FTR_RESIDENT_DATA, 0 );
    }
}



void HandleNotification( SysNotifyParamType* np )
{
    switch ( np->notifyType ) {
        case 'WiDn':
            InstallHede( true );
            break;
        case appCreator:
            InstallHede( false );
            break;
    }
}


UInt32 PilotMain( UInt16 command, void* cmdPBP, UInt16 flags )
{
    switch ( command ) {
        case sysAppLaunchCmdPanelCalledFromApp:
        case sysAppLaunchCmdNormalLaunch:
        {
            Err err;

            launchCode = command;

            ErrTry {
                err = errNone;
                if ( ApplicationStart() )
                    EventLoop();
                ApplicationStop();
            }
            ErrCatch( err ) {
                ApplicationStop();
                if ( err != errNone )
                    return err;
                break;
            } ErrEndCatch

            break;
        }

        case sysAppLaunchCmdSystemReset:
        case sysAppLaunchCmdSyncNotify: {
            PreferenceType p;

            LoadPrefs0( &p );

            InstallResident( p.active );
            break;
        }

        case sysAppLaunchCmdNotify: {
            HandleNotification( (SysNotifyParamType *)cmdPBP );
            break;
        }

        default:
            break;
    }
    return 0;
}

